(* exceptions *)

exception Fatal

exception NotFound
exception NoDropable
exception NoPickable

(* error report *)

let fatal (msg: string) = (print_string msg; raise Fatal)

let print_line (s: string) = Printf.printf ("%s\n") s; flush stdout

(* parsing *)

module S = Stream

let space = ' '

let ascii_of_0 = int_of_char '0'
let ascii_of_9 = int_of_char '9'

let is_digit c =
  let i = int_of_char c in
  (i >= ascii_of_0 && i <= ascii_of_9)

let ascii_of_A = int_of_char 'A'
let ascii_of_Z = int_of_char 'Z'

let is_empty (s: char S.t): bool =
  try S.empty s; true with S.Failure -> false

let read_int (s: char S.t): int =
  let rec aux (i) =
    if is_empty s then i
    else
      let c = S.next s in
      if (c = space) then i
      else
	if is_digit c then
	  aux (10 * i + int_of_char c - ascii_of_0)
	else fatal ("read_int: not a digit: " ^ Char.escaped c) in
  if is_empty s then fatal "read_int: empty streat" else aux 0

let read_ints (s: char S.t): int list =
  let rec aux (res: int list): int list =
    if is_empty s then List.rev res else aux (read_int s :: res) in
  aux []

let get_dimen (s: string): int * int = (* col number, row number *)
(*
  let _ = print_line ("get_dimen: " ^ s) in
*)
  let s = S.of_string s in
  let nc = read_int s in
  let nr = read_int s in
(*
  let _ = Printf.printf "nc = %i and nr = %i\n" nc nr; flush stdout in
*)
  (nc, nr)

let get_robspec (s: string): int * int * int = (* id, capacity and money *)
(*
  let _ = print_line ("get_robspec: " ^ s) in
*)
  let s = S.of_string s in
  let id = read_int s in
  let cap = read_int s in
  let money = read_int s in (id, cap, money)

type token =
  | N | S | E | W
  | P of int | D of int
  | X of int | Y of int
  | R of int

let token_pr = function
  | N -> "N"
  | S -> "S"
  | E -> "E"
  | W -> "W"
  | P id -> "P" ^ (string_of_int id)
  | D id -> "D" ^ (string_of_int id)
  | X id -> "X" ^ (string_of_int id)
  | Y id -> "Y" ^ (string_of_int id)
  | R id -> "R" ^ (string_of_int id)

let rec tokens_pr = function
  | [] -> ""
  | [t] -> token_pr t
  | t :: ts -> token_pr t ^ " " ^ tokens_pr ts

let tokenize (s: char S.t): token list =
  let rec aux (res: token list) =
    if (is_empty s) then (List.rev res)
    else
      let c = S.next s in
      match c with
      |	'#' -> aux (R (read_int s) :: res)
      |	'N' -> (S.junk s; aux (N :: res))
      |	'S' -> (S.junk s; aux (S :: res))
      |	'E' -> (S.junk s; aux (E :: res))
      |	'W' -> (S.junk s; aux (W :: res))
      |	'P' -> (S.junk s; aux (P (read_int s) :: res))
      |	'D' -> (S.junk s; aux (D (read_int s) :: res))
      |	'X' -> (S.junk s; aux (X (read_int s - 1) :: res))
      |	'Y' -> (S.junk s; aux (Y (read_int s - 1) :: res))
      |	_ -> fatal ("tokenize: unknown token: " ^ Char.escaped c ^ "\n") in
  aux ([])

(* 2-dimension array *)

module A = Array

type 'a array2 = ('a array) array

let get2 (a: 'a array2) (i: int) (j: int): 'a = A.get (A.get a i) j

let set2 (a: 'a array2) (i: int) (j: int) (x: 'a): unit =
  A.set (A.get a i) j x

let make2 (i: int) (j: int) (x: 'a): 'a array2 =
  let arr = A.make i (A.make j x) in
  for k = 1 to i - 1 do A.set arr k (A.make j x) done; arr


(* some datatypes *)

type entry =
    Wall
  | Water
  | Home
  | Space

let make_entry (c: char): entry =
  match c with
  | '#' -> Wall
  | '~' -> Water
  | '@' -> Home
  | '.' -> Space
  | _ -> fatal ("make_entry: " ^ Char.escaped c ^ "\n")

type gamemap = entry array2

type position = int * int

let is_neighbor ((x1, y1): position) ((x2, y2): position): bool =
  (x1 = x2 && (y1 = y2 + 1 || y1 = y2 - 1)) ||
  ((x1 = x2 + 1 || x1 = x2 - 1) && y1 = y2)

let string_of_position ((x, y): position): string =
  Printf.sprintf "(%i, %i)" x y

type spathmap = int array2

let print_row (row: int array) (n: int) =
  for i = 0 to n - 1 do
    print_int row.(i); print_string " ";
  done;
  print_newline ()

let print_spathmap (map: int array2) nr nc =
  for i = 0 to nr - 1 do
    print_row (map.(i)) nc
  done

type spathmaps = (spathmap option) array2

type pack = {
    p_id: int;
    p_dest: position;
    p_weight: int;
  } 

let pack_pr (pack: pack) =
  "pack(" ^ string_of_int pack.p_id ^ ")"

let rec packs_pr = function
  | [] -> ""
  | [pack] -> pack_pr pack
  | pack :: packs -> pack_pr pack ^ " " ^ packs_pr packs

let make_pack (id: int) (pos: position) (weight: int) = {
  p_id = id;
  p_dest = pos;
  p_weight = weight;
} 

let pack_find (packs: pack list) (pid: int): pack =
  let rec aux = function
    | [] -> raise NotFound
    | pack :: packs ->
	if pid = pack.p_id then pack else aux packs in
  aux packs

let pack_select (packs: pack list) (pid: int): pack * pack list =
  let rec aux acc = function
    | [] -> raise NotFound
    | pack :: rest ->
	if pid = pack.p_id then (pack, List.rev_append acc rest)
	else aux (pack :: acc) rest in
  aux [] packs

let pack_remove (packs: pack list) (pid: int): pack list =
  let rec aux acc = function
    | [] -> packs
    | pack :: rest ->
	if pid = pack.p_id then List.rev_append acc rest
	else aux (pack :: acc) rest in
  aux [] packs
  

type base = {
    b_pos: position;
    mutable b_packs: pack list;
  } 

let make_empty_base (pos: position) = {
  b_pos = pos;
  b_packs = [];
} 

type robot = {
    r_id: int;
    mutable r_pos: position;
  } 

let make_robot (rid: int) (rpos: position) =
(*
  print_line ("make_robot: " ^ string_of_int rid);
*)
  {
    r_id = rid; r_pos = rpos;
  } 

type myself = {
    my_id: int;
    my_cap: int;
    mutable my_money: int;
    mutable my_pos: position;
    mutable my_packs: pack list;
    mutable my_score: int;
    mutable my_weight: int;
  } 

(* robot commands *)

type direction = No | So | Ea | We

type action =
    Move of direction
  | Pick of int list
  | Drop of int list

let string_of_direction = function
    No -> "N"
  | So -> "S"
  | Ea -> "E"
  | We -> "W"

let rec string_of_ints = function
    [] -> ""
  | [n] -> string_of_int n
  | n :: ns -> string_of_int n ^ " " ^ string_of_ints ns

let string_of_action = function
    Move d -> "Move" ^ " " ^ string_of_direction d
  | Pick ids -> "Pick" ^ " " ^ string_of_ints ids
  | Drop ids -> "Drop" ^ " " ^ string_of_ints ids

(* global variables *)

let bases: base list ref = ref []
let bases2: base list ref = ref []

let bases_pr (bs: base list): string =
  let rec aux = function
    | [] -> ""
    | [b] -> string_of_position b.b_pos
    | b :: bs -> string_of_position b.b_pos ^ " " ^ aux bs in
  aux bs

let base_find (bs: base list) (pos: position): base option =
  let rec aux = function
    | [] -> None
    | b :: rest ->
	if b.b_pos = pos then Some b
	else aux rest in
  aux bs

let add_base (bases: base list ref) (b: base): unit =
(*
  print_line ("add_base: " ^ string_of_position b.b_pos)
*)
  match base_find !bases b.b_pos with
  | None -> bases := b :: !bases
  | Some _ -> ()


let base_remove (bases: base list ref) (pos: position): unit =
  let rec aux acc = function
    | [] -> ()
    | b :: rest ->
	if pos = b.b_pos then
	  bases := List.rev_append acc rest
	else aux (b :: acc) rest in
  aux [] !bases

let make_base (pos: position) (ns: int list): base =
  let rec aux packs = function
    | [] -> packs
    | id :: y :: x :: weight :: rest ->
	let pack = make_pack id (x-1, y-1) weight in
	aux (pack :: packs) rest
    | _ -> fatal ("make_base: ill format") in
  {
    b_pos = pos;
    b_packs = aux [] ns;
  } 

let other_robots: robot list ref = ref []

let add_robot (r: robot): unit =
  other_robots := r :: !other_robots

let robot_find (rid: int): robot option =
  let rec aux = function
    | [] -> None
    | r :: rest ->
	if rid = r.r_id then Some r
	else aux rest in
  aux !other_robots

let current_packs: pack list ref = ref []

(* networking *)

module U = Unix

let domain_of = function
  | U.ADDR_UNIX _ -> U.PF_UNIX
  | U.ADDR_INET _ -> U.PF_INET

let open_connection (sockaddr: U.sockaddr): in_channel * out_channel =
  let domain = domain_of sockaddr in
  let sock = U.socket domain U.SOCK_STREAM 0 in
  try
    U.connect sock sockaddr;
    (U.in_channel_of_descr sock, U.out_channel_of_descr sock)
  with exn -> U.close sock; raise exn    

let shutdown_connection (inchan: in_channel): unit =
  U.shutdown (U.descr_of_in_channel inchan) U.SHUTDOWN_SEND

let main_client (client_fun: in_channel -> out_channel -> 'a): unit =
  if A.length Sys.argv < 3
  then Printf.printf "usage: client server port\n"
  else
    let server = Sys.argv.(1) in
    let server_addr =
      try U.inet_addr_of_string server
      with
	Failure("inet_addr_of_string") ->
	  try (U.gethostbyname server).U.h_addr_list.(0)
	  with
	    Not_found ->
	      Printf.eprintf "%s: Unknown server\n" server; exit 2 in
    try
      let port = int_of_string (Sys.argv.(2)) in
      let sockaddr = U.ADDR_INET (server_addr, port) in
      let (ic, oc) = open_connection sockaddr in
      client_fun ic oc; shutdown_connection ic
    with
      Failure ("int_of_string") ->
	Printf.eprintf "bad port number"; exit 2  

(* decision *)

(* function for finding shortest path *)

module Q = Queue

let isW: entry -> bool = function
  | Wall -> true
  | Water -> true
  | _ -> false

let out_of_bounds (nr: int) (nc: int) ((x, y): position): bool = 
    (x < 0) || (y < 0) || (x >= nr) || (y >= nc)

let near_water gamemap nr nc (x, y): bool =
  let aux x y =
    if out_of_bounds nr nc (x, y) then false
    else
      match get2 gamemap x y with
      | Water -> true
      | _ -> false in
  if aux (x-1) y then true
  else if aux (x+1) y then true
  else if aux x (y-1) then true
  else if aux x (y+1) then true
  else false

let find_spathmap
      ((gamemap, (nr, nc)): gamemap * position) 
      ((dx, dy): position): spathmap = 
  let spathmap: spathmap = make2 nr nc (-1) in
  let pds: (position * int) Q.t = Q.create () in
  let rec bfs (): unit =
    if Q.length pds > 0 then
      let ((x, y), d) = Q.take pds in
      if out_of_bounds nr nc (x, y) then bfs ()
      else if isW (get2 gamemap x y) then bfs ()
      else if (get2 spathmap x y) <> -1 then bfs ()
      else begin
	set2 spathmap x y d;
	Q.add ((x - 1, y), d+1) pds; 
	Q.add ((x + 1, y), d+1) pds;
	Q.add ((x, y - 1), d+1) pds;
	Q.add ((x, y + 1), d+1) pds;
	bfs ()
      end in
  Q.add ((dx, dy), 0) pds; bfs ();
(*
  print_spathmap spathmap nr nc;
*)
  spathmap

(*
let find_shortest_path 
  (spathmaps: spathmaps)
  (gamemap: gamemap)
  (num_of_rows: int)
  (num_of_cols: int)
  ((source_row_index, source_col_index): position) 
  ((target_row_index, target_col_index): position) = 
  let 
    target_shortest_path_matrix =
      match get2 spathmaps target_row_index target_col_index with
      | Some matrix -> matrix
      | None -> 
          let 
            matrix = 
              find_spathmap
                (gamemap, (num_of_rows, num_of_cols)) 
                (target_row_index, target_col_index)
          in
            set2 spathmaps target_row_index target_col_index (Some matrix);
            matrix
  in 
    get2 target_shortest_path_matrix 
         source_row_index 
         source_col_index
*)

let pick_from (packs: pack list) (myself: myself): pack list =
    let left_cap = ref (myself.my_cap - myself.my_weight) in
    let rec aux_same_dest (pickeds: pack list) (left: pack list) 
    (packs: pack list): pack list * pack list =
        let dests = List.map (fun pack -> pack.p_dest) myself.my_packs in
        match packs with 
        | [] -> (pickeds, left) 
        | pack :: rest -> 
                if (pack.p_weight <= !left_cap) && (List.mem pack.p_dest dests) 
                then aux_same_dest (pack :: pickeds) left rest
                else aux_same_dest pickeds (pack :: left) rest in
    let (pickeds1, left) = aux_same_dest [] [] packs in 
    let rec aux (pickeds: pack list) (packs: pack list): pack list =
        match packs with 
        | [] -> pickeds
        | pack :: rest ->
                if pack.p_weight <= !left_cap 
                then let _ = (left_cap := !left_cap - pack.p_weight) in
                     aux (pack :: pickeds) rest
                else aux pickeds rest in
    aux pickeds1 left

let find_base
    (gamemap: gamemap) (spathmaps: spathmaps)
    (posList: position list)
    (pos: position) (nr: int) (nc: int): position * int =
    let init_shortest = (0, 0) in 
    let rec aux (shortest: position * int) (posList: position list): position * int =
        match posList with
        | [] -> shortest
        | posit :: rest ->
                let (row, col) = posit in
                let (myrow, mycol) = pos in 
                begin
                    match get2 spathmaps row col with
                    | None -> 
                            let matrix = find_spathmap
                            (gamemap,(nr, nc)) posit in
                            let _ = set2 spathmaps row col (Some matrix) in
                            let path_length = get2 matrix myrow mycol in
                            if (path_length < snd shortest) ||
                            (snd shortest = -1) then 
                                aux (posit, path_length) rest
                            else aux shortest rest
                    | Some matrix -> 
                            let path_length = get2 matrix myrow mycol in
                            if (path_length < snd shortest) || 
                            (snd shortest = -1) then 
                                aux (posit, path_length) rest 
                            else aux (posit, path_length) rest
                end in
    aux (init_shortest, -1) posList

let find_path
    (spathmaps: spathmaps) (my_pos: position) (dest: position) 
    (pathlen: int) (nr: int) (nc: int) : action =
    let (dx, dy) = dest in
    let (x, y) = my_pos in
    let cond spathmap x y =
      if out_of_bounds nr nc (x, y) then false
      else if get2 spathmap x y = pathlen - 1 then true
      else false in
    match get2 spathmaps dx dy with
    | None -> fatal "find_path"
    | Some spathmap ->
        if cond spathmap (x - 1) y then Move So
        else if cond spathmap (x + 1) y then Move No
        else if cond spathmap x (y + 1) then Move Ea
	else if cond spathmap x (y - 1) then Move We
        else (print_spathmap spathmap nr nc; fatal ("find_path: impossible"))

let mk_decision (gamemap: gamemap) (nr: int) (nc: int)
                (spathmaps: spathmaps) (myself: myself): (int * action) = 
     let bid_move =
       if near_water gamemap nr nc myself.my_pos then 10 else 1 in
     let bid_pick = 10 in
     let bid_drop = 10 in
     let dropable_packs: pack list =
       List.filter 
         (function pack -> pack.p_dest = myself.my_pos) 
         myself.my_packs in
     match dropable_packs with
     | [] -> begin
         let pickedList = pick_from !current_packs myself in
         match pickedList with
         | [] ->
	     let posList =
	       match myself.my_packs with
	       | [] -> begin
		   match !bases with
		   | [] -> begin
		       match !bases2 with
		       | [] -> []
		       | _ -> List.map (function base -> base.b_pos) !bases2
		   end
		   | _ -> List.map (function base -> base.b_pos) !bases
	       end
	       | _ -> List.map (fun pack -> pack.p_dest) myself.my_packs in
	     begin
	       match posList with
	       | [] -> (bid_pick, Pick [])
               | _ ->
		   let (pos, pathlen) =
		     find_base
		       gamemap spathmaps posList myself.my_pos nr nc in
		   (bid_move,
		    find_path spathmaps myself.my_pos pos pathlen nr nc)
	     end
         | _ -> (bid_pick,
		 Pick (List.map (fun pack -> pack.p_id) pickedList))
     end
     | _ ->
	 let pidList =
	   List.map (function pack -> pack.p_id) dropable_packs in
         (bid_drop, Drop pidList)

(* the main function *)

let client_fun (ic: in_channel) (oc: out_channel): 'a =
  try
    let _ = output_string oc ("Player\n") in
    let _ = flush oc in

    let (nc, nr) = get_dimen (input_line ic) in

    let gamemap = make2 nr nc Wall in
    let initialize_row (i: int): unit =
      let s = input_line ic in
(*
      print_line s;
*)
      for j = 0 to nc - 1 do
	let entry = make_entry (String.get s j) in
	match entry with
	| Home ->
	    add_base bases (make_empty_base (i, j));
	    set2 gamemap i j entry
	| _ -> set2 gamemap i j entry
      done in
    let _ =
      for i = 0 to nr - 1 do initialize_row (i) done in

    let spathmaps: spathmaps = make2 nr nc None in

    let myself =
      let (id, cap, money) = get_robspec (input_line ic) in
      {
	my_id = id;
	my_cap = cap;
	my_money = money;
	my_pos = (0, 0);
	my_packs = [];
	my_score = 0;
	my_weight = 0;
      }	in

    let initialize_robots =
      let s = input_line ic in
(*
      let _ = print_line ("initialize_robots: " ^ s) in
*)
      let s = S.of_string s in
      let tokens = tokenize s in
      let rec aux = function
	| [] -> ()
	| R (id) :: X(y) :: Y(x) :: rest ->
	    if id = myself.my_id then aux rest
	    else begin
	      add_robot (make_robot id (x, y));
	      aux rest
	    end
	| _ -> fatal ("aux: wrong format") in
      aux tokens in

    let update_bases () =
      let s = input_line ic in
(*
      let _ = print_line ("update_bases: " ^ s) in
*)
      let s = S.of_string s in
      let ns = read_ints s in
      match ns with
      |	[] ->
	  base_remove bases myself.my_pos;
	  base_remove bases2 myself.my_pos;
	  current_packs := []
      | _ -> 
	  let b = make_base myself.my_pos ns in
	  current_packs := b.b_packs;
	  add_base bases b in

    let update_robots () =
      let s = input_line ic in
(*
      let _ = print_line ("update_robots: " ^ s) in
*)
      let s = S.of_string s in
      let tokens = tokenize s in
      let rec aux (rs: robot list) (rid: int) (tokens: token list): unit =
	if rid = myself.my_id then aux_myself rs tokens
	else
	  match robot_find rid with
	  | None -> begin
	      match tokens with
	      |	X (y) :: Y (x) :: rest ->
		  let r = make_robot rid (x, y) in
		  aux_other rs r rest
	      |	_ -> fatal ("update_robots: unknown robot: " ^
			    tokens_pr tokens)
	  end
	  | Some r -> aux_other rs r tokens

      and aux_myself (rs: robot list): token list -> unit = function
	| [] -> other_robots := List.rev rs
	| X (y) :: Y (x) :: rest -> aux_myself rs rest
	| P id :: rest ->
	    let (pack, packs) = 
	      try pack_select !current_packs id
	      with NotFound ->
		fatal ("update_robots: aux_myself: P: pack not found") in
	    myself.my_packs <- pack :: myself.my_packs;
	    myself.my_weight <- myself.my_weight + pack.p_weight;
	    aux_myself rs rest

	| D id :: rest ->
	    let (pack, packs) =
	      try pack_select myself.my_packs id
	      with NotFound ->
		fatal ("update_robots: aux_myself: D: pack not found") in
	    myself.my_packs <- packs;
	    myself.my_weight <- myself.my_weight - pack.p_weight;
	    if pack.p_dest = myself.my_pos then
	      myself.my_score <- myself.my_score + pack.p_weight
	    else
	      add_base
		bases {b_pos = myself.my_pos; b_packs = [pack]};
	    aux_myself rs rest

	| N :: rest -> 
	    let (x, y) = myself.my_pos in
	    myself.my_pos <- (x+1, y); aux_myself rs rest
	| S :: rest -> 
	    let (x, y) = myself.my_pos in
	    myself.my_pos <- (x-1, y); aux_myself rs rest
	| E :: rest -> 
	    let (x, y) = myself.my_pos in
	    myself.my_pos <- (x, y+1); aux_myself rs rest
	| W :: rest -> 
	    let (x, y) = myself.my_pos in
	    myself.my_pos <- (x, y-1); aux_myself rs rest

	| R rid :: rest -> aux rs rid rest
	| tokens -> fatal ("aux_myself: " ^ tokens_pr tokens)

      and aux_other (rs: robot list) (r: robot): token list -> unit = function
	| [] -> other_robots := List.rev (r :: rs)
	| X (y) :: Y (x) :: rest ->
	    r.r_pos <- (x, y); aux_other rs r rest
	| P id :: rest -> begin
	    match base_find !bases (r.r_pos) with
	    | None -> aux_other rs r rest
	    | Some b ->
		b.b_packs  <- pack_remove b.b_packs id;
		aux_other rs r rest
	end

	| D id :: rest ->
	    add_base bases2 {b_pos = r.r_pos; b_packs = []};
	    aux_other rs r rest

	| N :: rest -> 
	    let (x, y) = r.r_pos in
	    r.r_pos <- (x+1, y); aux_other rs r rest
	| S :: rest -> 
	    let (x, y) = r.r_pos in
	    r.r_pos <- (x-1, y); aux_other rs r rest
	| E :: rest -> 
	    let (x, y) = r.r_pos in
	    r.r_pos <- (x, y+1); aux_other rs r rest
	| W :: rest -> 
	    let (x, y) = r.r_pos in
	    r.r_pos <- (x, y-1); aux_other rs r rest

	| R rid :: rest -> aux (r :: rs) rid rest
	| tokens -> fatal ("aux_other: " ^ tokens_pr tokens) in
      match tokens with
      |	R (rid) :: rest -> aux [] rid rest
      |	_ -> fatal "update_robots: empty tokens" in

    while true do
(*
      print_line ("my position: " ^ string_of_position myself.my_pos);
      print_line ("my score: " ^ string_of_int myself.my_score);
      print_line ("my money: " ^ string_of_int myself.my_money);
*)
      update_bases ();

(*
      print_line ("bases: " ^ bases_pr !bases);
      print_line ("bases2: " ^ bases_pr !bases2);
      print_line ("current_packs: " ^ packs_pr !current_packs);
*)

      let (bid, act) = mk_decision gamemap nr nc spathmaps myself in
      myself.my_money <- myself.my_money - bid;
(*
      print_line ("action: " ^ string_of_action act);
*)
      output_string
	oc (string_of_int bid ^ " " ^ string_of_action act ^ "\n");
      flush oc;
      update_robots ();
    done;

    shutdown_connection ic
  with
  | Exit -> exit 0
  | exn -> shutdown_connection ic; exit 1

let go_client (): unit = main_client client_fun

let _ = go_client ()
